#!/usr/bin/perl -w
# Journal & report module
#
# Copyright (C) 2007 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF with some parts
# derived from lsb-autotest written by Jiri Dluhos <jdluhos@suse.cz>
# Copyright (C) 2005-2007 SuSE Linux Products GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

sub BEGIN {
    $| = 1; # Immediate flush of all output
}

use strict;
use IO::File;

my $result_dir;
my $result_subdir;
my $output_file_name;           # -o <filename>
my $app_name = "";              # -n <appname>

# HTML colors used in the report.
my $dark_gray = "rgb(64, 64, 64)";
my $gray = "rgb(200, 200, 200)";
my $neutral_gray = "rgb(230, 230, 230)";
my $bright_gray = "rgb(240, 240, 240)";
my $fluorescent_green = "rgb(64, 255, 64)";
my $fluorescent_yellow = "rgb(255, 255, 64)";
my $fluorescent_red = "rgb(255, 96, 96)";
my $pale_blue = "rgb(220, 220, 240)";
my $dark_blue = "rgb(0, 0, 128)";
my $dark_red = "rgb(128, 0, 0)";
my $red = "rgb(255, 0, 0)";
my $green = "rgb(0, 160, 0)";
my $blue = "rgb(0, 0, 255)";

# Statistics for each test
my %stats_failures_reported = ();
my %stats_warnings_reported = ();
my %stats_fips_reported = ();
my %stats_notices_reported = ();
my %stats_tests_total = ();
my %stats_tests_passed = ();
my %stats_tests_skipped = ();
my %stats_tests_quelled = ();
my %stats_tests_expected = ();
my %stats_tests_skips_unexpected = ();

# Summary statistics
my $tests_total = 0;
my $tests_passed = 0;
my $tests_quelled = 0;
my $tests_skipped = 0;

my $tests_cancelled = 0;
my $failures_reported = 0;
my $overall_verdict = "";

# Host information
my $host_machine;
my $host_name;
my $host_platform;
my $host_os;
my $host_kernel;
#my $etc_issue;
my $lsb_version;
my $test_comments;

# The problem database, hashed by the "test_name test_point" strings.
my %problem_database;

# The final exit code, which is 0 if all tests have passed, or 1 if some
# failures were detected.
my $exit_code = 0;

# The report file.
my $report;

# This variable is set to 1 if part of the problem table was already printed
# (at least the header), false if not (this implies that no problems were
# found so far).
my $problem_table_printed = 0;

# A monotonously increasing, unique identifier of a message box.
my $msgbox_id = 1;

# Copy of the input line that was processed last.
my $line = "";

#---------------------------------------------------------------------

# The journal file being processed.
my $journal_file;
my $journal_name;

# The title of the journal being processed.
my $journal_title = "";
my @journal_titles = ();

# The HTML anchors are used for marking this journal in the resulting report.
# They are unique for each journal, so we use them as journal IDs.
my @journal_anchors_sorted = ();
my %journal_title_from_anchor = ();

# Reports for each test suite
my %reports = ();  # $reports{$journal_title} = $html_report_text

sub report {
    $_ = shift;
    $reports{$journal_title} .= $_;
}

#---------------------------------------------------------------------

# Map of TET result codes to human-readable severity names.
my %result_code_map = (
    0, "Passed",
    1, "Failed",
    2, "Unresolved",
    3, "Unused",
    4, "Unsupported",
    5, "Untested",
    6, "Uninitiated",
    7, "Unreported",
    101, "Warning",
    102, "FIP",
    103, "Unimplemented",
    104, "Unapproved",
);

#---------------------------------------------------------------------

sub html_entities {
    my $text = shift;

    $text =~ s/(^[\n\r\s]+|[\n\r\s]+$)//g;
    # Escape HTML entities
    $text =~ s/<br \/>/\n/gi;
    $text =~ s/&/&amp;/g;
    $text =~ s/</&lt;/g;
    $text =~ s/>/&gt;/g;
    $text =~ s/[\n\r]+/<br \/>\n/g;

    return $text;
}
#---------------------------------------------------------------------
sub fail {
    my $text = shift;

    print STDERR "$text\n";
    exit(-1);
}

#---------------------------------------------------------------------
# Prints time and date in a human-readable format
sub format_time {
    my $timestamp = shift;

    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday) =
        localtime ($timestamp);

    return sprintf ("%02d:%02d:%02d %02d-%02d-%04d\n", $hour, $min, $sec,
        $mday, ($mon + 1), ($year + 1900) );
}

#---------------------------------------------------------------------

# Processes the result of a test point.
#
# Arguments:
# $1 - test number (0 if the tests are not numbered)
# $2 - test name
# $3 - test point (0 if the tests are not divided into test points)
# $4 - severity
# $5 - test messages
sub process_test_point_result {
    my ($test_number, $test_point, $test_name, $severity, $test_messages) = @_;

    # Sanity check: Test name must be valid.
    unless ($test_name) {
        print ("Empty test name\n");
        return;
    }

    # Compare severity strings in lowercase.
    my $lowercase_severity = lc($severity);

    # Increment the total test count.
    $stats_tests_total{$journal_title}++;

    # Passed and unused tests are not printed, only counted.
    if ($lowercase_severity eq "passed") {
        $stats_tests_passed{$journal_title}++;
        return;
    }
    elsif (($lowercase_severity eq "unused") || ($lowercase_severity eq "untested") ||
        ($lowercase_severity eq "unsupported") || ($lowercase_severity eq "unimplemented")) {
        $stats_tests_skipped{$journal_title}++;
        return;
    }

    my $quelled = 0;
    my $resolution = "";
    my $comment = "";
    my $lsb_report_id = "";
    my $lsb_resolution_id = "";

    # Print the problem to the table.
    print_problem_table_line(
        $test_number,
        $test_point,
        $test_name,
        $severity,
        $resolution,
        $comment,
        $lsb_report_id,
        $lsb_resolution_id,
        $test_messages
    );
}
#---------------------------------------------------------------------

# Prints a header of the problem table. It is called automatically
# from report_problem() when the first problem is reported.
# (If there are no problems during the whole test run, the table
# is not printed at all.)
sub print_problem_table_header() {

    report("<h4>Problem Summary</h4>\n\n");
    report("<p>Click on lines in the table to see the details " .
        "about each problem.</p>\n\n");
    report("<table width=\"100%\">\n");
    report("<thead>\n");
    report("<tr style=\"font-weight:bold; background-color:$gray;\">\n");
    report("<td>Test Name</td><td>Severity</td><td>Resolution</td>\n");
    report("</tr>\n");
    report("</thead>\n");
    report("<tbody>\n");
}

#---------------------------------------------------------------------

# Prints a line of the problem table.
#
# Arguments:
#
# $1 - test number (0 if the tests are not numbered, or if this problem line
#      is not connected to any particular test).
# $2 - test point number (0 if there is no test point).
# $3 - test name (empty if this problem line is not connected to any
#      particular test).
# $4 - severity.
# $5 - resolution (empty if unknown).
# $6 - comment
# $7 - report ID in the LSB database (empty if not yet reported).
# $8 - resolution ID in the LSB database (empty if not yet resolved).
# $9 - messages printed by the testing program.
sub print_problem_table_line
{
    my $test_number = shift;
    my $test_point = shift;
    my $test_name = shift;
    my $severity = shift;
    my $resolution = shift;
    my $comment = shift;
    my $lsb_report_id = shift;
    my $lsb_resolution_id = shift;
    my $test_messages = shift;

    # Compare the severity string case-insensitively (we are only humans).
    my $lowercase_severity = lc($severity);

    # Choose the color for the specified message severity.
    my $color;
    if ($lowercase_severity =~ /\(quelled|waived\)/) {
        # Anything what is quelled is harmless, so use neutral gray
        # for all quelled problems.
        $color = "gray";
    }
    elsif (($lowercase_severity eq "failed") || ($lowercase_severity eq "failure") ||
        ($lowercase_severity eq "unresolved") || ($lowercase_severity eq "unreported") ||
        ($lowercase_severity eq "uninitiated") || ($lowercase_severity eq "test suite error")
    ) {
        # Use fluorescent red for various failures.
        $color = "red";

        # Increment the number of failures reported.
        $stats_failures_reported{$journal_title}++;

        # Even a single failure in a single test is enough for flipping
        # the exit code from 0 to 1.
        $exit_code = 1 << 8;
    }
    elsif ($lowercase_severity eq "warning") {
        # Use fluorescent yellow for warnings.
        $color = "yellow";

        # Increment the number of warnings reported.
        $stats_warnings_reported{$journal_title}++;
    }
    elsif ($lowercase_severity eq "fip") {
        # Use fluorescent green for FIPs.
        $color = "green";

        $severity .= " [<a href='tests_help.pl#FIP' title='Help'>?</a>]";

        # Increment the number of FIPs reported.
        $stats_fips_reported{$journal_title}++;
    }
    elsif ($lowercase_severity eq "notice") {
        # Use neutral gray for harmless messages.
        $color = "gray";

        # Increment the number of notices reported.
        $stats_notices_reported{$journal_title}++;
    }
    else {
        print ("Unrecognized severity code '$severity'\n");
    }

    # If the problem table header was not yet printed, print it now.
    unless ($problem_table_printed) {
        print_problem_table_header();
        $problem_table_printed = 1;
    }

    # Give the message box a unique ID.
    my $message_box_id = "msgbox" . $msgbox_id; $msgbox_id++;

    # Start the line (with attached script that opens the detail block
    # on clicking).
    report("<tr class=\"ptl_$color\" title='Click here!' onclick=\"flip('$message_box_id');\">");

    # Print the test name and test point (if available).
    report("<td>");
    if ($test_name) {
        report("$test_name ");
    }
    else {
        report("(none) ");
    }
    if ($test_point > 0) {
        report("<span style=\"color:$dark_blue;\">$test_point</span>\n");
    }
    report("</td>\n");

    # Print severity and resolution.
    report("<td>$severity</td>\n");
    report("<td>$resolution</td>\n");

    # Finish the line.
    report("</tr>\n");

    # Print the detail block, which is initially hidden and unrolls when
    # the table row is clicked (via the javascript above).
    report("<tr id=\"$message_box_id\" style=\"display:none;\">\n");
    report("<td colspan=\"3\">");
    report("<div style='padding-left:32px;'>");
    report("Messages from the test:\n");
    report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
    if ($test_messages) {
    	my $details_link;
    	$test_messages = html_entities($test_messages);
    	if ($test_messages =~ s/Symbol (.*) is used, but is not included in/Symbol <b>$1<\/b> is used, but is not included in/) {
    		$details_link = "<br />See details about the interface in the <a href=\"http://www.linux-foundation.org/navigator/browse/int_single.php?cmd=list-by-name&amp;Iname=$1\" target=\"_blank\">LSB Database Navigator</a>";
    	}
    	else {
    		$details_link = "";
    	}
        report("<p style='font-family:monospace'>".$test_messages.$details_link."</p>");
    } else {
        report("(none)");
    }
    report("\n</div>\n");
    if ($comment) {
        report("Comment:\n");
        report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
        report("$comment\n</div>\n");
    }
    if ($lsb_report_id or $lsb_resolution_id) {
        report("LSB Report Status:\n");
        report("<div style=\"color:$dark_gray;padding-left:32px;\">\n");
        if ($lsb_report_id) {
            report("Reported to the LSB Committee as ");
            report("<a href=\"https://www.linux-foundation.org/cert/pr.php?view=$lsb_report_id\">");
            report("PR $lsb_report_id</a>.<br />\n");
        }
        if ($lsb_resolution_id) {
            report("Resolved (resolution ID is ");
            report("<span style=\"color:$dark_blue;\">$lsb_resolution_id</span>).<br />\n");
        }
        report("</div>");
    }
    report("</div>");
    if ( $color eq "red" ) {
         my $esc_journal_title = $journal_title;
         $esc_journal_title =~ s/ /%20/g;
         my $esc_test_case = $test_name.' '.$test_point;
         $esc_test_case =~ s/ /%20/g;
         report("<div align=\"right\">"
            ."<input type=\"button\" value=\"Report Bug in the Test\""
            ." onclick=\"javascript:bug_report('$result_dir/$output_file_name', '$esc_journal_title', '$esc_test_case');\""
            ." title=\"Press this button if you beleive the problem is in the test, not in your system.\" /></div>"
            );
    }
    report("</td>");
    report("</tr>\n");
}

#---------------------------------------------------------------------

# Prints the footer of the problem table. Does nothing if the problem table
# was not printed at all.
sub finish_problem_table() {

    # Complain if we have not seen any tests at all; this may mean the test suite
    # terminated somehow at the very start.
    if ($stats_tests_total{$journal_title} == 0) {
        print_problem_table_line(
            0,
            0,
            "(none)",
            "Failure",
            "No tests found in the journal",
            "",
            "",
            $line
        );
    }

    # Close the table or report that no problems were detected
    if ($problem_table_printed) {
        report("</tbody>\n");
        report("</table>\n");
    }
    else {
        report("<h4>Problem Summary</h4>\n");
        report("<p><font color='#006600'><b>No problems were detected.</b></font></p>\n");
    }
}

#---------------------------------------------------------------------

# Starts a two-column table.
sub start_two_column_table {
    report("<table>\n");
    report("<tbody>\n");
}

#---------------------------------------------------------------------

# Prints a single line of a two-column table.
#
# Arguments:
# $1 - The first column.
# $2 - The second column.
# $3 (optional) - The foreground color of the second column.
# $4 (optional) - The background color of the second column.
sub two_column_table_line {
    my $first_column = shift;
    my $second_column = shift;
    my $foreground_color = shift;
    my $background_color = shift;

    # Build the style specifier for the second column.
    my $second_column_style = "padding: 2px; margin: 2px;";

    # If the second column contains a number, print it right-aligned.
#    if ($second_column =~ /^[0-9\.]*$/) {
        $second_column_style .= "text-align: right; ";
#    }

    # Apply the foreground and background colors if specified.
    if ($foreground_color) {
        $second_column_style .= "color: $foreground_color; ";
    }
    if ($background_color) {
        $second_column_style .= "background-color: $background_color; ";
    }
    else {
        $second_column_style .= "background-color: $neutral_gray; ";
    }

    # Print the line with appropriate decorations.
    return "<tr style=\"padding: 4px; margin: 4px;\">".
    "<td style=\"background-color:$gray;\">$first_column</td>\n".
    "<td style=\"$second_column_style\">$second_column</td>\n".
    "</tr>\n";
}
#---------------------------------------------------------------------

# Prints a two-column table line to the test report
sub print_two_column_table_line {
    report(two_column_table_line(@_));
}
#---------------------------------------------------------------------

# Finishes a two-column table.
sub finish_two_column_table {
    report("</tbody>\n");
    report("</table>\n");
}

#---------------------------------------------------------------------

sub print_statistics_line {
    my ($title, $value, $warn_if_nonzero, $warning_color) = @_;

    # Use neutral gray background by default.
    my $background_color = $neutral_gray;

    # If a warning is requested, change the color
    # when the value is nonzero.
    if ($warn_if_nonzero && ($value > 0) ) {
        print_two_column_table_line($title, $value, $dark_blue,
            $warning_color);
    }
    else {
        print_two_column_table_line($title, $value, $dark_blue);
    }
}


#---------------------------------------------------------------------

# Prints the statistics.
sub print_statistics() {
    report("<h4>Statistics</h4>\n");

    # Print the statistics table.
    my $background_color;
    start_two_column_table();
    print_statistics_line("Total tests", $stats_tests_total{$journal_title}, 0);
    if ($stats_tests_expected{$journal_title} > 0) {
        print_statistics_line("Total tests expected", $stats_tests_expected{$journal_title},
            $stats_tests_total{$journal_title}<$stats_tests_expected{$journal_title}?1:0, $fluorescent_yellow );
    }
    print_statistics_line("Passed tests", $stats_tests_passed{$journal_title}, 0);
    if ($stats_tests_skipped{$journal_title} > 0) {
        print_statistics_line("Tests not relevant/not applicable", $stats_tests_skipped{$journal_title}, 0);
    }
    if ($stats_tests_skips_unexpected{$journal_title} > 0) {
        print_statistics_line("Test skips unexpected", $stats_tests_skips_unexpected{$journal_title}, 1, $fluorescent_yellow);
    }
    if ($stats_tests_quelled{$journal_title} > 0) {
        print_statistics_line("Quelled problems", $stats_tests_quelled{$journal_title}, 0);
    }
    if ($stats_fips_reported{$journal_title} > 0) {
        print_statistics_line("FIPs", $stats_fips_reported{$journal_title}, 1, $fluorescent_green);
    }
    if ($stats_warnings_reported{$journal_title} > 0) {
        print_statistics_line("Warnings", $stats_warnings_reported{$journal_title}, 1, $fluorescent_yellow);
    }
    if ($stats_failures_reported{$journal_title} > 0) {
        print_statistics_line("Failures", $stats_failures_reported{$journal_title}, 1, $fluorescent_red);
    }

    # Give the final verdict.
    if ($stats_failures_reported{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Failure");
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        print_two_column_table_line("Verdict", "Warnings");
    }
    else {
        print_two_column_table_line("Verdict", "Success");
    }

    # Finish the table.
    finish_two_column_table();

    # Sanity check: If everything looks fine, check if the number of tests
    # matches the sum of skipped, passed, quelled and FIPed tests. This helps
    # preventing false negatives (which are most dangerous).
    if ($stats_failures_reported{$journal_title} == 0) {
        if ($stats_tests_skipped{$journal_title} + $stats_tests_passed{$journal_title} + $stats_fips_reported{$journal_title}
            + $stats_tests_quelled{$journal_title} != $stats_tests_total{$journal_title}) {

            # Oops.
            print ("No failures reported but some tests are missing\n");
        }
    }
}

#---------------------------------------------------------------------

my %journal_tree;

sub add_elem_to_subtree($$$$) {
	my ($subtree, $elem_name, $f_type, $j_anchor) = @_;
	$elem_name =~ s/^\/*([^\/].*)$/$1/;
	my $idx = index($elem_name, "/");
	if ($idx > 0) {
		my $d_name = substr($elem_name, 0, $idx);
		$elem_name = substr($elem_name, $idx + 1);
		$subtree->{"children"}->{$d_name}->{"type"} = "dir";
		$subtree->{"children"}->{$d_name}->{"journal_anchor"} = undef;
		&add_elem_to_subtree($subtree->{"children"}->{$d_name}, $elem_name, $f_type, $j_anchor);
	}
	else {
		$subtree->{"children"}->{$elem_name}->{"children"} = undef;
		$subtree->{"children"}->{$elem_name}->{"type"} = $f_type;
		$subtree->{"children"}->{$elem_name}->{"journal_anchor"} = $j_anchor;
	}
}

sub add_elem_to_tree($$$$$) {
	my ($tree_root, $elem_name, $f_type, $j_anchor) = @_;
	if (!$tree_root) {
		if (!$journal_tree{$elem_name}) {
	   		$journal_tree{$elem_name}->{"children"} = undef;
	   	}
   		$journal_tree{$elem_name}->{"type"} = $f_type;
   		$journal_tree{$elem_name}->{"journal_anchor"} = $j_anchor;
   	}
   	else {
   		if (!$journal_tree{$tree_root}) {
    		if ($tree_root =~ m/^PKG: /) {
	    		$journal_tree{$tree_root}->{"type"} = "pkg";
	    		$journal_tree{$tree_root}->{"journal_anchor"} = undef;
    		}
    		elsif ($tree_root =~ m/\.rpm$/) {
	    		$journal_tree{$tree_root}->{"type"} = "pkg";
	    		$journal_tree{$tree_root}->{"journal_anchor"} = undef;
    		}
    		else {
	    		$journal_tree{$tree_root}->{"type"} = "dir";
	    		$journal_tree{$tree_root}->{"journal_anchor"} = undef;
    		}
   		}
   		&add_elem_to_subtree($journal_tree{$tree_root}, $elem_name, $f_type, $j_anchor);
   	}
}

sub parse_journal
{
    $journal_name = shift;
    my $descr_filename = $journal_name.".INFO";
    my $title = "";
    my $anchor = "";
    my $description = "";

    # Parse INFO file
    my $res;
    if ( !-f $descr_filename || !open(FILE, "$descr_filename") ) {
        print ("Couldn't open file '$descr_filename' for reading.\n");
        $anchor = $journal_name;
        $anchor =~ s/.*\///;
        $anchor =~ s/\..*$//;
        $title = $anchor;
    } else {
        while ( my $line = <FILE> ) {
            if ( $line =~ /^\s*TITLE:\s+(.*)$/ ) { $title = $1; }
            elsif ( $line =~ /^\s*ANCHOR:\s+(.*)$/ ) { $anchor = $1; }
            elsif ( $line =~ /^\s*JOURNAL_FILENAME:\s+(.*)$/ ) {
                #$journal_name = $1;
            }
        }
        close FILE;
    }

    unless ($journal_name) {
        print "Syntax error in $descr_filename"
             ." (JOURNAL_FILENAME: $journal_name)\n";
        return 1;
    }

    $journal_title = $title;
    if ($journal_title =~ m/^\|(.*)$/) {
    	add_elem_to_tree(undef, $1, "file", $anchor, $journal_title);
    }
    elsif ($journal_title =~ m/^([^\|]+\.rpm)\|$/) {
    	add_elem_to_tree(undef, $1, "pkg", $anchor, $journal_title);
    }
    elsif ($journal_title =~ m/^([^\|]+)\|(.+)$/) {
    	add_elem_to_tree($1, $2, "file", $anchor, $journal_title);
    }

    # Store the journal title and anchor.
    push @journal_titles, $journal_title;
    $journal_title_from_anchor{$anchor} = $journal_title;

    $problem_table_printed = 0;

    $stats_failures_reported{$journal_title} = 0;
    $stats_warnings_reported{$journal_title} = 0;
    $stats_fips_reported{$journal_title} = 0;
    $stats_notices_reported{$journal_title} = 0;
    $stats_tests_total{$journal_title} = 0;
    $stats_tests_passed{$journal_title} = 0;
    $stats_tests_skipped{$journal_title} = 0;
    $stats_tests_quelled{$journal_title} = 0;
    $stats_tests_expected{$journal_title} = 0;
    $stats_tests_skips_unexpected{$journal_title} = 0;

    # If journal isn't closed then close it.
    if ( $journal_file ) { close_journal (); }

    # Open the journal file.
    print ("Reading journal: $journal_name\n");
    $journal_file = new IO::File;
    unless (open ($journal_file, $journal_name)) {

        # The journal file could not be open; mention this in the log
        # (so that we know that something was missing) and return.

        my $journal_title_disp = $journal_title;
        $journal_title_disp =~ s/^\|//;
        $journal_title_disp =~ s/\|$//;
        $journal_title_disp =~ s/\|/ =&gt; /;

        $tests_cancelled++;

        report( "<p><font color='#A00000'><b>No result.</b></font>"
            ." $journal_title_disp failed to run, interrupted"
            ." or crashed during the execution."
        );
        print "Failed to read journal: $journal_name\."
            ."\n$journal_title_disp failed to run, interrupted or crashed"
            ." during the execution.\n($journal_name was not present).\n";
        $journal_file = undef;
    } else {
        if ( $description ) {
            report("<p>".html_entities($description)."</p>");
        }
        report("<p><a href=\"get.pl$result_dir/$journal_name\" target=\"_blank\">Open the journal file</a></p>" );
    }

    parse_tet_journal();
}

#---------------------------------------------------------------------

sub close_journal
{
    unless ($journal_file) { return; }

    # Finish the problem table.
    finish_problem_table();

    # Print the statistics.
    print_statistics();

    # Close the journal file.
    $journal_file->close ();

    # Reset state variables.
    $journal_file = undef;
    $journal_title = "";
}

#---------------------------------------------------------------------

# The main TET journal parser function which loads the TET journal
# and produces a HTML report. Uses stdin if the input file name is "-".
#
# Parameters:
#    $1 - the title to be printed before the report (typically the name
#         of the test the journal belongs to).
#    $2 - a HTML anchor to produce (useful for referencing to the results
#         of the test in a larger document).
#    $3 - the file name of the journal.
sub parse_tet_journal
{
    unless ($journal_file) { return; }

    # This flag is set to nonzero if we are inside a test point.
    my $inside_test = 0;
    my $inside_test_point = 0;

    # Configuration variables (from lines with code 30)
    my %config_vars = ();

    # Read the journal line by line.
    my ($test_name, $test_number, $test_point, $test_messages);
    my $test_points_in_this_case;
    my $line_number = 0;
    while ($line = <$journal_file>) {
        $line_number++;
        # Each journal line consists of three parts, separated by
        # a vertical bar:
        #
        # 1. A numeric line code, determining the meaning of the line.
        # 2. Additional data, format depends on line code.
        # 3. Additional text (message for the user, or some comment).
        unless ($line =~ /^(\d*)\|(.*)\|(.*)$/) {

            # Lines that do not conform to this format are reported
            # to the user and skipped.
            print_problem_table_line(
                0,
                0,
                "(none)",
                "Warning",
                "Unrecognized entry in test journal",
                "$journal_name($line_number)",
                "",
                "",
                $line
            );
            next;
        }
        my $line_code = $1;
        my $line_data = $2;
        my $line_text = $3;

        # Determine the type of the line and process it appropriately.
        if ($line_code == 0) {
            # Code 0 -- Test Case Controller Start.
            # The line structure is:
            # 0|version time date|who
            unless ($line_data =~ /^(\S*)\s*(\S*)\s(\S*)$/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # This line should be one of the first;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case Start entry appears in the middle of the journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $test_engine_version = $1;
                my $test_start_time = $2;
                my $test_start_date = $3;
                my $test_command_line = $line_text;

                # Print the test engine version, time and date to the output.
                report("<h4>Test Information</h4>");
                start_two_column_table();
                print_two_column_table_line("Test Engine Version", $test_engine_version);
                print_two_column_table_line("Test Executed At", "$test_start_time $test_start_date");
                print_two_column_table_line("Test Command Line", "$test_command_line");
                finish_two_column_table();
            }
        }
        elsif ($line_code == 5) {

            # Code 5 -- Local System Information.
            # The line structure is:
            # 5|sysname nodename release #date machine|text
            unless ($line =~ /^5\|(\S*)\s*(\S*)\s*(\S*)\s*#(.*)\s(\S+)\|/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # The system information should be one of the first items;
            # if it comes in the middle of normal results, it is strange.
            if ($problem_table_printed) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "System info entry appears in the middle of the journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    "$line_data\n$line_text"
                );
            }
            else {
                my $system_name = $1;
                my $node_name = $2;
                my $release = $3;
                my $date = $4;
                my $machine = $5;

                # Print the local system data.
                report("<h4>Tested System Information</h4>\n");
                start_two_column_table();
                print_two_column_table_line("System Name", $system_name);
                print_two_column_table_line("Node Name", $node_name);
                print_two_column_table_line("Release", $release);
                print_two_column_table_line("Release Date", $date);
                print_two_column_table_line("Machine", $machine);
                finish_two_column_table();
            }
        }
        elsif ($line_code == 10) {

            # Code 10 -- Test Case Start.
            # The line structure is:
            # 10|activitynum testname hh:mm:ss|list of invocable components
            unless ($line =~ /^\d*\|(\d*)\s*(.*)\s+\S+\|/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            if ($test_name) {
                # The previous test case was not properly finished;
                # issue a warning.
                print_problem_table_line(
                    $1,
                    0,
                    $test_name,
                    "Warning",
                    "Test case starts before the previous one ($test_name) has finished",
                    "$journal_name($line_number)",
                    "",
                    "",
                    ""
                );
            }

            # Store the test name and number.
            $test_number = $1;
            $test_name = $2;
            $test_name =~ s|^.*/([^/]+)$|$1|;
            $test_name =~ s|^(.*)-pass2$|$1|;

            # Reset the test point number and counter.
            $test_point = 0;
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 15) {
            # Code 15 -- Test Case Manager Start.
            # Not interesting to us.
        }
        elsif ($line_code == 20) {
            # Code 20 -- Local/Remote System Configuration Start.
            # Not interesting to us.
        }
        elsif ($line_code == 30) {

            # Code 30 -- Configuration Variables.
            # The line structure is:
            # 30||VARIABLE=value
            if ($line =~ /^30\|\|([^=]+)=(.*)/) {
                my $varname = $1;
                my $value = $2;

                $config_vars{$varname} = $value;
            }
        }
        elsif ($line_code == 40) {
            # Code 40 -- Configuration End.

            # Print the local system data.
            report("<h4>Configuration information</h4>\n");
            start_two_column_table();
            for my $varname qw(VSX_NAME LSB_VERSION LSB_MODULES TET_VERSION XT_FONTPATH) {
                if ( exists $config_vars{$varname} ) {
                    print_two_column_table_line($varname, $config_vars{$varname});
                }
            }
            finish_two_column_table();
        }
        elsif ($line_code == 50) {
            # Code 50 -- Test Case Controller Message.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Test case controller reports a problem",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 70) {
            # Code 70 -- Scenario Information.

            # LSB-runtime-test only:
            # for the .os sections need to double expected counts to allow
            # for the macro versions.
                if ( $line =~ /^70\|.*total tests in (ANSI|POSIX|LSB|PTHR).os ([0-9]+)/ ) {
                    $stats_tests_expected{$journal_title} += $2;
                    $stats_tests_expected{$journal_title} += $2;
                }
                elsif ( $line =~ /^70\|.*total tests in ([A-Za-z.0-9_-]+) ([0-9]+)/ ) {
                    $stats_tests_expected{$journal_title} += $2;
                }
                elsif ( $line =~ /total\s+tests\s+in/ ) {
                    complain ("$journal_name($line_number): A suspicious record in journal.")
                }
        }
        elsif ($line_code == 80) {
            # Code 80 -- Test Case End.
            unless ($test_name) {
                # Test case end found when no test case was started;
                # issue a warning.
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Test Case End entry found when no test case seems to be running",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $test_messages
                );
            }

            # Warn if the test case seems to be empty.
            if ($test_points_in_this_case == 0) {
                print_problem_table_line(
                    $test_number,
                    0,
                    $test_name,
                    "Warning",
                    "Test case contains no test points",
                    "",
                    "",
                    "",
                    ""
                );
            }

            # Clear the test name, number, and messages.
            $test_name = "";
            $test_number = 0;
            $test_messages = "";
            $test_points_in_this_case = 0;
        }
        elsif ($line_code == 90) {
            # Code 90 -- User Abort.
            print_problem_table_line(
                0,
                0,
                "(none)",
                "Test Suite Error",
                "Test suite execution interrupted by the user",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 100) {
            # Code 100 -- Captured Output.
            # Not interesting to us.
        }
        elsif ($line_code == 110) {
            # Code 110 -- Build Start.
            # Not interesting to us.
        }
        elsif ($line_code == 130) {
            # Code 130 -- Build End.
            # Not interesting to us.
        }
        elsif ($line_code == 200) {
            # Code 200 -- Test Purpose Start.
            # The structure of the line is as follows:
            # 200|activitynum pointnum hh:mm:ss|comment
            unless ($line =~ /^\d*\|(\d*)\s*(\d*)/) {
                print_problem_table_line(
                    0,
                    0,
                    "(none)",
                    "Warning",
                    "Syntax error in test journal",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $line);
                next;
            }

            # Check if the previous test point was properly finished.
            if ($test_point > 0) {
                if ($test_point == $2) {
                    # Some test suites have a strange habit of re-starting
                    # the same test point more than once.
                    print_problem_table_line(
                        $test_number,
                        $test_point,
                        $test_name,
                        "Notice",
                        "Test was restarted in the middle",
                        "",
                        "",
                        "",
                        ""
                    );
                }
                else {
                    # Notify about a disappeared test.
                    print_problem_table_line(
                        $test_number,
                        $2,
                        $test_name,
                        "Warning",
                        "Test point starts before the previous one ($test_point) has finished",
                        "",
                        "",
                        "",
                        ""
                    );
                }
            }

            # Clear the messages.
            $test_messages = "";

            # Store the test point number.
            $test_point = $2;

            # Increment the number of test points.
            $test_points_in_this_case++;
        }
        elsif ($line_code == 220) {
            # Code 220 -- Test Purpose Result.
            # The structure of the line is as follows:
            # 220|activitynum pointnum result time|comment
            unless ($line =~ /^\d*\|(\d*)\s*(\d*)\s*(\d*)/) {
                complain("Syntax error in journal");
                next;
            }

            # Print a warning if the test purpose seems to never been started.
            if ($test_point == 0) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Warning",
                    "Test result entry found but the test point had not started",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }

            # Print a warning if the test purpose number is different from
            # what we expected.
            elsif ($test_point != $2) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Warning",
                    "Test purpose number at the end ($2) is different " .
                    "from the starting one ($test_point)",
                    "",
                    "",
                    "",
                    $test_messages
                );
            }

            # Get the result code.
            my $result_code = $3;

            # Find the meaning of the result code (severity).
            my $severity = $result_code_map{$result_code};
            unless ($severity) {
                print_problem_table_line(
                    $test_number,
                    $test_point,
                    $test_name,
                    "Failure",
                    "Unrecognized test result code ($result_code)",
                    "$journal_name($line_number)",
                    "",
                    "",
                    $test_messages
                );
            }
            else {
                # Process the result.
                process_test_point_result(
                    $test_number,
                    $test_point,
                    $test_name,
                    $severity,
                    $test_messages
                );
            }

            # Clear the test point number.
            $test_point = 0;
        }
        elsif ($line_code == 310) {
            # 310 -- Clean Start.
            # Not interesting to us.
        }
        elsif ($line_code == 320) {
            # 320 -- Clean End.
            # Not interesting to us.
        }
        elsif ($line_code == 400) {
            # 400 -- Invocable Component Start.
            # Not interesting to us.
        }
        elsif ($line_code == 410) {
            # 410 -- Invocable Component End.
            # Not interesting to us.
        }
        elsif ($line_code == 510) {
            # 510 -- Error reported by the underlying test suite.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Test Suite Error",
                "Error reported by the test suite",
                "",
                "",
                "",
                $line_text
            );
        }
        elsif ($line_code == 520) {
            # 520 -- Message from the underlying test suite.

            # Coalesce consequent rows of spaces in the message
            # to a single space. This improves matching with markers.
            my $message = $line_text;
            $message =~ s/\s+/ /g;

            # Add the message to the accumulated messages.
            $test_messages .= $message."\n";
        }
        elsif ($line_code == 600) {
            # 600 -- Parallel Start.
            # Not interesting to us.
        }
        elsif ($line_code == 620) {
            # 620 -- Parallel End.
            # Not interesting to us.
        }
        elsif ($line_code == 630) {

            # 630 -- Implied Sequential Start.
            # Not interesting to us.
        }
        elsif ($line_code == 640) {
            # 640 -- Implied Sequential End.
            # Not interesting to us.
        }
        elsif ($line_code == 700) {
            # 700 -- Repeat Start.
            # Not interesting to us.
        }
        elsif ($line_code == 720) {
            # 720 -- Repeat End.
            # Not interesting to us.
        }
        elsif ($line_code == 730) {
            # 730 -- Timed Loop Start.
            # Not interesting to us.
        }
        elsif ($line_code == 740) {
            # 740 -- Timed Loop End.
            # Not interesting to us.
        }
        elsif ($line_code == 750) {
            # 750 -- Random Start.
            # Not interesting to us.
        }
        elsif ($line_code == 760) {
            # 760 -- Random End.
            # Not interesting to us.
        }
        elsif ($line_code == 800) {
            # 800 -- Remote Start.
            # Not interesting to us.
        }
        elsif ($line_code == 820) {
            # 820 -- Remote End.
            # Not interesting to us.
        }
        elsif ($line_code == 830) {
            # 830 -- Distributed Start.
            # Not interesting to us.
        }
        elsif ($line_code == 840) {
            # 840 -- Distributed End.
            # Not interesting to us.
        }
        elsif ($line_code == 900) {
            # 900 -- Test Case Controller End.
            # Not interesting to us.
        }
        else {
            # Other lines are notified to the user and otherwise ignored.
            print_problem_table_line(
                $test_number,
                $test_point,
                $test_name,
                "Notice",
                "Unrecognized journal entry (code $line_code)",
                "$journal_name($line_number)",
                "",
                $line
            );
        }
    }

    # Finish the journal processing and close it.
    close_journal ();
}

#---------------------------------------------------------------------

sub max(@) {
	if (scalar(@_) == 0) {
		return undef;
	}
	my $res = $_[0];
	foreach (@_) {
		$res = $_ if ($res < $_);
	}
	return $res;
}

# Returns array with number of passes, failes and warnings in the test.
# The test is specified via unique HTML anchor.
sub report_testcases_info($) {
    my $journal_anchor = $_[0];
    if (!$journal_anchor) {
    	return (-1, -1, -1);
    }
	if (!$journal_title_from_anchor{$journal_anchor}) {
		return (-1, -1, -1);
	}
    my $journal_title = $journal_title_from_anchor{$journal_anchor};
	return ($stats_tests_passed{$journal_title}, $stats_failures_reported{$journal_title}, $stats_warnings_reported{$journal_title});
}

my $separator_line = "<img src=\"images/environment-spacer.gif\" width=\"100%\" height=\"1\" style=\"background-color: white;\" alt=\"\" />";
my @verdicts = ("&nbsp;", "&nbsp;", "Success", "Warnings", "Failed");
my @verdict_colors = ("#e7e7e7", "#d0d0ff", $fluorescent_green, $fluorescent_yellow, $fluorescent_red);

sub report_failures_verdict($$) {
    my ($journal_anchor, $is_root) = @_;

    if (!$journal_anchor) {
    	return ($is_root ? 1 : 0); # Nothing
    }

	if (!$journal_title_from_anchor{$journal_anchor}) {
    	return ($is_root ? 1 : 0); # Nothing
	}

    $journal_title = $journal_title_from_anchor{$journal_anchor};

    if ($stats_failures_reported{$journal_title} > 0) {
        return 4; # Failed
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        return 3; # Warning
    }
    elsif ($stats_tests_passed{$journal_title} == 0) {
        return 4; # Failed
    }
    elsif (($stats_tests_expected{$journal_title} > 0) and
           ($stats_tests_expected{$journal_title} > $stats_tests_total{$journal_title})) {
        return 3; # Warning
    }
    else {
        return 2; # Success
    }
}

sub report_failures_cell($$) {
    my ($journal_anchor, $is_root) = @_;
    my $bg_color = ($is_root ? "#d0d0ff" : "#e7e7e7");

    if (!$journal_anchor) {
    	return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
    }

	if (!$journal_title_from_anchor{$journal_anchor}) {
    	return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}

    $journal_title = $journal_title_from_anchor{$journal_anchor};

    my $status = "";
    if ($stats_failures_reported{$journal_title} > 0) {
        $bg_color = $fluorescent_red;
        $status = "Failures: ".$stats_failures_reported{$journal_title};
    }
    elsif ($stats_warnings_reported{$journal_title} > 0) {
        $bg_color = $fluorescent_yellow;
        $status = "Warnings: ".$stats_warnings_reported{$journal_title};
    }
    elsif ($stats_tests_passed{$journal_title} == 0) {
        $bg_color = $fluorescent_red;
        $status = "<span title=\"'No result' means the tests failed"
        ." to run, interrupted or crashed during the execution.\">No Result"
        ."&nbsp; <a href=\"tests_help.pl#No_result\" title=\"Help\"><img src=\"images/question.png\" alt=\"[?]\" width=\"16\" height=\"17\" /></a></span>";
    }
    elsif ($stats_tests_expected{$journal_title} > 0
          && $stats_tests_expected{$journal_title} > $stats_tests_total{$journal_title} )
    {
        $bg_color = $fluorescent_yellow;
        $status = "Warning";
    }
    else {
        $bg_color = $fluorescent_green;
        $status = "Success";
    }

    return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />$status<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line";
}

sub report_passes_cell($$) {
    my ($journal_anchor, $is_root) = @_;
    my $bg_color = ($is_root ? "#d0d0ff" : "#e7e7e7");
    my $passes = (report_testcases_info($journal_anchor))[0];
    if ($passes == -1) {
    	return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
    }
    else {
	    return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: $passes<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line";
    }
}

my $failures_cell = "";
my $passes_cell = "";

sub print_tree_tab($$$) {
	my ($j_tree, $parent_id, $tab) = @_;
	my %sum_my_info = ("passes" => 0, "test_status" => 0, "passes_cell" => "", "failures_cell" => "");
	my @children = sort keys %$j_tree;
	for (my $i = 0; $i <= $#children; $i++) {
		my $elem_name = $children[$i];
		my $elem_id = "$parent_id:$elem_name";
		my $img_f_type;
		my $img_plus_minus;
		my $img_suffix;
		my $tab_new;
		if ($i < $#children) {
			$img_suffix = "d";
			$tab_new = "$tab<img src=\"images/tree_d.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}
		else {
			$img_suffix = "";
			$tab_new = "$tab<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}
		
		if ($j_tree->{$elem_name}{"type"} eq "dir") {
			$img_f_type = "<img src=\"images/tree_dir.png\" alt=\"[D]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_minus_u$img_suffix.png\" alt=\"[+]\" width=\"21\" height=\"21\" id=\"icon$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '_u$img_suffix');\" />";
		}
		else {
			$img_f_type = "<img src=\"images/tree_file.png\" alt=\"[F]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_dr$img_suffix.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}

		my $elem_link = $elem_name;
		if ($j_tree->{$elem_name}{"journal_anchor"}) {
			$elem_link = "<a href=\"#".$j_tree->{$elem_name}{"journal_anchor"}."\" title=\"Click here to proceed to details\">$elem_link</a>";
			push @journal_anchors_sorted, $j_tree->{$elem_name}{"journal_anchor"};
		}
		$report->print("<div style=\"background-color: #e7e7e7;\">$tab$img_plus_minus$img_f_type<b>$elem_link</b><img src=\"images/environment-spacer.gif\" width=\"5\" height=\"21\" alt=\"\" /><br /></div>$separator_line");

		# Calculate passes and failures for the element and its children
		my $local_passes_cell;
		my $local_failures_cell;
		my ($local_test_passes, $local_test_failures, $local_test_warnings) = report_testcases_info($j_tree->{$elem_name}{"journal_anchor"});
		my $local_test_status;
		if ($local_test_passes == -1) {
			$local_passes_cell = "<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_failures_cell = "<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_test_passes = 0;
			$local_test_status = 0;
		}
		else {
			$local_passes_cell = report_passes_cell($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$local_test_status = report_failures_verdict($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$local_failures_cell = report_failures_cell($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$sum_my_info{"passes"} += $local_test_passes;
			$sum_my_info{"test_status"} = max($local_test_status, $sum_my_info{"test_status"});
		}
		if ($j_tree->{$elem_name}{"children"}) {
			$report->print("<div id=\"root$elem_id\">\n");
			my %sum_children_info = &print_tree_tab($j_tree->{$elem_name}{"children"}, $elem_id, $tab_new);
			$report->print("</div>\n");
			$local_test_passes += $sum_children_info{"passes"};
			$sum_my_info{"passes"} += $local_test_passes;
			$local_passes_cell = "<div style=\"display: none;\" id=\"pass-total$elem_id\">\n<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: ".$local_test_passes."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div id=\"pass-children$elem_id\">\n".$local_passes_cell.$sum_children_info{"passes_cell"}."</div>\n";
			$sum_my_info{"passes_cell"} .= $local_passes_cell;

			$local_test_status = max($local_test_status, $sum_children_info{"test_status"});
			$sum_my_info{"test_status"} = max($local_test_status, $sum_my_info{"test_status"});
			$local_failures_cell = "<div style=\"display: none;\" id=\"fail-total$elem_id\">\n<div style=\"background-color: ".$verdict_colors[$local_test_status].";\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />".$verdicts[$local_test_status]."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div id=\"fail-children$elem_id\">\n".$local_failures_cell.$sum_children_info{"failures_cell"}."</div>\n";
			$sum_my_info{"failures_cell"} .= $local_failures_cell;
		}
		else {
			$sum_my_info{"passes_cell"} .= $local_passes_cell;
			$sum_my_info{"failures_cell"} .= $local_failures_cell;
		}
	}
	return %sum_my_info;
}

sub print_tree($) {
	my $j_tree = $_[0];
	$report->print("<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\"><tr><td><img src=\"images/environment-spacer.gif\" width=\"2\" height=\"1\" alt=\"\" /></td><td valign=\"top\">\n");
	foreach my $elem_name (sort keys %$j_tree) {
		my $elem_id = $elem_name;
		$elem_id =~ s/\//:/g;
		$elem_id =~ s/^PKG: /:PKG::/;
		my $img_f_type;
		my $img_plus_minus;
		if ($j_tree->{$elem_name}{"type"} eq "dir") {
			$img_f_type = "<img src=\"images/tree_dir.png\" alt=\"[D]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_plus.png\" alt=\"[+]\" width=\"21\" height=\"21\" id=\"icon$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '');\" />";
		}
		elsif ($j_tree->{$elem_name}{"type"} eq "pkg") {
			$img_f_type = "<img src=\"images/tree_pkg.png\" alt=\"[P]\" width=\"21\" height=\"21\" />";
			if ($j_tree->{$elem_name}{"children"}) {
				$img_plus_minus = "<img src=\"images/tree_plus.png\" alt=\"[+]\" width=\"21\" height=\"21\" id=\"icon$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '');\" />";
			}
			else {
				$img_plus_minus = "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
			}
		}
		else {
			$img_f_type = "<img src=\"images/tree_file.png\" alt=\"[F]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}
		my $elem_link = $elem_name;
		if ($j_tree->{$elem_name}{"journal_anchor"}) {
			$elem_link = "<a href=\"#".$j_tree->{$elem_name}{"journal_anchor"}."\" title=\"Click here to proceed to details\">$elem_link</a>";
			push @journal_anchors_sorted, $j_tree->{$elem_name}{"journal_anchor"};
		}
		$report->print("<div style=\"background-color: #d0d0ff;\">\n$img_plus_minus$img_f_type<b>$elem_link</b><img src=\"images/environment-spacer.gif\" width=\"5\" height=\"21\" alt=\"\" /><br />\n</div>$separator_line");

		# Calculate passes and failures for the element and its children
		my $local_passes_cell;
		my $local_failures_cell;
		my ($local_test_passes, $local_test_failures, $local_test_warnings) = report_testcases_info($j_tree->{$elem_name}{"journal_anchor"});
		my $local_test_status;
		if ($local_test_passes == -1) {
			$local_passes_cell = "<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_failures_cell = "<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_test_passes = 0;
			$local_test_status = 1;
		}
		else {
			$local_passes_cell = report_passes_cell($j_tree->{$elem_name}{"journal_anchor"}, 1);
			$local_test_status = report_failures_verdict($j_tree->{$elem_name}{"journal_anchor"}, 1);
			$local_failures_cell = report_failures_cell($j_tree->{$elem_name}{"journal_anchor"}, 1);
		}
		if ($j_tree->{$elem_name}{"children"}) {
			$report->print("<div id=\"root$elem_id\" style=\"display: none;\">\n");
			my %sum_children_info = &print_tree_tab($j_tree->{$elem_name}{"children"}, $elem_id, "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />");
			$report->print("</div>\n");
			$local_test_passes += $sum_children_info{"passes"};
			$local_passes_cell = "<div id=\"pass-total$elem_id\">\n<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: ".$local_test_passes."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div style=\"display: none;\" id=\"pass-children$elem_id\">\n".$local_passes_cell.$sum_children_info{"passes_cell"}."</div>\n";
			$local_test_status = max($local_test_status, $sum_children_info{"test_status"});
			$local_failures_cell = "<div id=\"fail-total$elem_id\">\n<div style=\"background-color: ".$verdict_colors[$local_test_status].";\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />".$verdicts[$local_test_status]."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div style=\"display: none;\" id=\"fail-children$elem_id\">\n".$local_failures_cell.$sum_children_info{"failures_cell"}."</div>\n";
		}
		$passes_cell .= $local_passes_cell;
		$failures_cell .= $local_failures_cell;
	}
	$report->print(<<DATA);
</td>
<td><img src="images/environment-spacer.gif" width="1" height="1" alt="" /></td>
<td valign="top" style="white-space: nowrap;">
$failures_cell
</td>
<td><img src="images/environment-spacer.gif" width="1" height="1" alt="" /></td>
<td valign="top" style="white-space: nowrap;">
$passes_cell
</td>
</tr></table>
DATA
}

sub compile_report {
    # Print Title
    $report->print ("<h1>LSB Test Results for <font color=\"blue\">".($app_name ? $app_name : join(", ", @journal_titles))."</font> on <font color=\"blue\">$host_platform</font></h1>\n");
    $report->print (<<DATA);
The complete console output log is available <a href="get.pl$result_dir/test_logs/tests_output" target="_blank">here</a>.<br /><br />

Extended information about the application tested is also available.
<ul>
  <li><a href="tests_results.pl?details=$result_subdir&amp;analysis=1">View the list of external libraries and interfaces</a> required by the application.</li>
  <li><b>Note:</b> You can <a href="upload_data.pl?details=$result_subdir"><b>send this information</b></a> to the Linux Foundation for publishing in the Linux Applications section of the <a href="http://www.linux-foundation.org/navigator/browse/app.php" target="_blank">LSB Navigator</a>.</li>
</ul>
<br />

<h3>Test Info</h3>
DATA

    # Print general information
    $report->print (
    "<table>\n"
    .two_column_table_line("Host Name", $host_name)
    .two_column_table_line("Host OS", $host_os)
    .two_column_table_line("Host Machine", $host_machine)
    .two_column_table_line("Host Platform", $host_platform)
    .two_column_table_line("Host Kernel", $host_kernel)
#    .two_column_table_line("/etc/issue", "<pre>$host_etc_issue</pre>")
    ."</table>\n"
    );

    # Print the index tree
    $report->print("<h3>Tests Executed</h3>\n");

	print_tree(\%journal_tree);

    $report->print(<<DATA); # DO NOT BREAK THE SPAN LINE
<br />
<span id="cert-integration" style="display:none">FILE=$result_dir/$output_file_name</span>
DATA

    $report->print("<br />\n");
    
    # Test journals
    foreach my $journal_anchor (@journal_anchors_sorted) {
    	if ($journal_title_from_anchor{$journal_anchor}) {
    		$journal_title = $journal_title_from_anchor{$journal_anchor};
    	}
    	else {
    		next;
    	}
        my $test_report = ($reports{$journal_title} or "");
        # Print the title.
        $journal_title =~ s/^\|//;
        $journal_title =~ s/\|$//;
        $journal_title =~ s/\|/ =&gt; /;
        $report->print("<hr />\n");
        $report->print("<h2><a name=\"".$journal_anchor."\"></a><font color=\"#336699\">"
        		.$journal_title."</font> &nbsp; <font size='-1' color='#888888'></font></h2>\n");

        $report->print($test_report);
    }
}

#---------------------------------------------------------------------

sub sum {
    my $sum = 0;
    foreach (@_) { $sum += $_; }
    return $sum;
}

#---------------------------------------------------------------------

sub generate_INFO_file {
    my $INFO_file_name = "INFO";

    my $file;
    unless (open($file, "> $INFO_file_name")) {
        fail("Couldn't open file '$INFO_file_name' for writing.");
    }

    print $file ""
        .$tests_total." Tests Total\n"
        .$tests_passed." Tests Passed\n"
        .$failures_reported." Failures\n"
        .$tests_quelled." Tests Quelled\n"
        .$tests_skipped." Tests Skipped\n"
        ;

    if ( $tests_cancelled > 0 ) {
        print $file "There are cancelled tests\n";
    }

    print $file "Verdict: $overall_verdict\n";

    print $file "Application: $app_name\n";

    print $file "\n".$test_comments;

    close($file);
}

#---------------------------------------------------------------------

sub Init {
    # Open the report files.
    $report = new IO::File;
    unless ($report->open ("> $result_dir/test_log.html")) {
        fail ("Could not open the report file for writing");
    }
    $report->print ("<html>\n");
    $report->print ("
<head>
<style type=\"text/css\">
.ptl_gray { background: #E6E6E6;}
.ptl_gray:hover { background: #D6D6F0;}
.ptl_green { background: #CDFFCD;}
.ptl_green:hover { background: #A0FFA0;}
.ptl_yellow { background: #FFFFCD;}
.ptl_yellow:hover { background: #FFFF90;}
.ptl_red { background: #FFCDCD;}
.ptl_red:hover { background: #FFA0A0;}
.ptl_gray,.ptl_green,.ptl_yellow,.ptl_red { cursor: pointer; }
</style>
</head>
    ");
    $report->print ("<body>\n");
    $report->print (<<DATA);

<script language="javascript" type="text/javascript">
// <![CDATA[
// Preload some images that are not initially visible
var preload_image_names = new Array('tree_d.png', 'tree_dir.png', 'tree_dr.png', 'tree_drd.png', 'tree_file.png', 'tree_minus.png', 'tree_minus_u.png', 'tree_minus_ud.png', 'tree_n.png', 'tree_pkg.png', 'tree_plus.png', 'tree_plus_u.png', 'tree_plus_ud.png');
var preload_images = new Array();
for (var i=0; i<preload_image_names.length; ++i) {
	preload_images[i] = new Image;
	preload_images[i].src = 'images/' + preload_image_names[i];
}

function show_hide(id, icon_suffix) {
	var elem = document.getElementById('root' + id);
	var icon = document.getElementById('icon' + id);
	var passTotal = document.getElementById('pass-total' + id);
	var passChildren = document.getElementById('pass-children' + id);
	var failTotal = document.getElementById('fail-total' + id);
	var failChildren = document.getElementById('fail-children' + id);
	if (elem && icon && passTotal && passChildren && failTotal && failChildren) {
		var to_show = (elem.style.display == 'none');
		if (to_show) {
			elem.style.display = '';
			passTotal.style.display = 'none';
			passChildren.style.display = '';
			failTotal.style.display = 'none';
			failChildren.style.display = '';
			icon.setAttribute('src', 'images/tree_minus' + icon_suffix + '.png');
		}
		else {
			elem.style.display = 'none';
			passTotal.style.display = '';
			passChildren.style.display = 'none';
			failTotal.style.display = '';
			failChildren.style.display = 'none';
			icon.setAttribute('src', 'images/tree_plus' + icon_suffix + '.png');
		}
	}
}

function flip (message_box_id) {
    msgbox = document.getElementById(message_box_id);
    if (msgbox.style.display != 'none') {
        msgbox.style.display = 'none';
    }
    else {
        msgbox.style.display = '';
    }
}

function bug_report(journal_path, test_app) {
    if (confirm('A new E-mail to lf_lsbcert\@lists.linux-foundation.org will now be created.\\nPlease, specify all the necessary details and attach the journal files:\\n\\n' + journal_path + '\\n\\nContinue?'))
    {
        document.location = 'mailto:lf_lsbcert\@lists.linux-foundation.org?Subject=Problem%20report:%20lsbappchk%20/%20' + test_app;
    }
}
// ]]>
</script>

DATA
}

#---------------------------------------------------------------------

sub Finish {

    # Close the report files.
    $report->print ("</body>\n</html>\n");
    $report->close ();

    return $exit_code;
}

#---------------------------------------------------------------------

# Read and return file contents
sub read_file {
    my $filename = shift;
    my $text = "";

    unless (open FILE, $filename) {
        print ("Can't open $filename\n");
        return "";
    }

    while (<FILE>) {
        $text .= $_;
    }
    close FILE;

    return $text;
}

sub read_test_info {
	$host_machine       = `uname -m`; $host_machine =~ s/[\r\n]//g;
	$host_name          = `uname -n`; $host_name =~ s/[\r\n]//g;
    $host_platform      = read_file("$result_dir/test_logs/host_platform");

    # The OS version is determined via the lsb_release tool
    # (this should be the canonical, distro independent way).
    $host_os = `lsb_release -s -d`;
    $host_os =~ s/\"(.*)\"/$1/;    # remove quotes
    $host_os =~ s/[\r\n]//g;       # remove trailing whitespace

	$host_kernel = `uname -r -v`;
	$host_kernel =~ s/[\r\n]//g;

    # Read the /etc/issue file, if present. Sometimes it is the only place
    # where an exact OS version can be found.
    #$etc_issue = `cat /etc/issue`;

    $lsb_version        = read_file("$result_dir/test_logs/lsb_version");
    $test_comments      = read_file("$result_dir/test_logs/test_comments");
}

#---------------------------------------------------------------------

# Generates a HTML report and detects test failures.
# Returns the return code of the reporting utility, which is 0 if all tests
# have passed, 1 if some tests have failed.
sub generate_report
{
    print ("Generating HTML report\n");

    #Go to the output directory
    chdir($result_dir);

    # Read the test information.
    read_test_info();

    $result_subdir = $result_dir;
    $result_subdir =~ s/^.*\///;

    unless ($output_file_name) {
        $output_file_name = "lsbappchk-$result_subdir";
    }
    if ($output_file_name !~ /\./) {
        $output_file_name .= ".tgz";
    }

    # Open a report file and print a header
    Init();

    # Parse each test journal
    opendir DIR, "test_journals" or die "Can't open directory $result_dir/test_journals: $!\n";
    my @files = readdir(DIR);
    foreach my $file (@files) {
        if ( $file =~ m/^(.*\.journal)\.INFO$/ ) {
            parse_journal("test_journals/$1");
        }
    }

    # Sum up statistics
    $failures_reported = sum( values %stats_failures_reported );
    $tests_total = sum( values %stats_tests_total );
    $tests_passed = sum( values %stats_tests_passed );
    $tests_quelled = sum( values %stats_tests_quelled );
    $tests_skipped = sum( values %stats_tests_skipped );

    # Give the final verdict
    if ( ($failures_reported > 0) || ($tests_cancelled > 0) ) {
        $overall_verdict = "Failed";
    }
    else {
        $overall_verdict = "Success";
    }

    # Write INFO file
    generate_INFO_file ();

    compile_report ();

    if ( -e "$result_dir/$output_file_name" ) {
        unlink("$result_dir/$output_file_name");
    }
    chdir("$result_dir/..");
    my $output_file_name_esc = $output_file_name;
    $output_file_name_esc =~ s/ /\\ /g;
    my $result_subdir_esc = $result_subdir;
    $result_subdir_esc =~ s/ /\\ /g;
    system ("tar czf $output_file_name_esc $result_subdir_esc && mv $output_file_name_esc $result_subdir_esc");

    my $exit_code = Finish();
    if ($exit_code == 0) {
        # Report was generated, no failures were detected.
        return 0;
    }
    elsif (($exit_code >> 8) == 1) {
        # Report was generated, some test failures were detected.
        return 1;
    }
    else {
        fail ("Error generating report (exit code $exit_code)");
    }
}

sub read_command_line
{
    my $argc = scalar(@ARGV);
    if ($argc < 1) {
        print "Usage: $0 <test_journals_dir>\n";
        return 1;
    }
    $result_dir = $ARGV[0];
    # Make $result_dir path absolute
    if ($result_dir !~ /^\//)
    {
        my $working_directory = `pwd`;
        chomp ($working_directory);
        $result_dir = $working_directory."/".$result_dir;
    }
    $result_dir =~ s/[\/\.]+$//; # remove trailing /.

    my $output_file_name_expected = 0;
    my $app_name_expected = 0;

    for (my $i = 0; $i<$argc; $i++) {
        my $arg = $ARGV[$i];

        if ($output_file_name_expected) {
            $output_file_name_expected = 0;
            $output_file_name = $arg;
        }

        if ($app_name_expected) {
            $app_name_expected = 0;
            $app_name = $arg;
        }

        if ($arg eq "-o") {
            $output_file_name_expected = 1;
        }
        elsif ($arg eq "-n") {
            $app_name_expected = 1;
        }
    }
}

### BEGIN ###

read_command_line();

$exit_code = generate_report();

exit($exit_code);

### END ###
